/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.internal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.conqat.lib.commons.collections.IdentityHashSet;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.IDecoration;
import org.eclipse.jface.viewers.ILabelProviderListener;
import org.eclipse.jface.viewers.ILightweightLabelDecorator;
import org.eclipse.jface.viewers.LabelProviderChangedEvent;
import org.eclipse.ui.progress.UIJob;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation.ESeverity;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IConstraintCheckerService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.listener.IPersistencyServiceListener;
import org.fortiss.tooling.kernel.ui.ESharedImages;
import org.fortiss.tooling.kernel.ui.internal.views.MarkerViewPart;
import org.fortiss.tooling.kernel.ui.service.IMarkerService;

/**
 * This class implements the {@link IMarkerService} interface.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: A7423A2812630B2E50752906688D3A1C
 */
public class MarkerService implements IMarkerService,
		IPersistencyServiceListener, ILightweightLabelDecorator {

	/** ID of this decorator. */
	public static final String ID = "org.fortiss.tooling.kernel.ui.internal.MarkerService";

	/** Stores the violations for each {@link ITopLevelElement}. */
	private final HashMap<ITopLevelElement, CacheEntry> violationCache = new HashMap<ITopLevelElement, CacheEntry>();

	/** List of listeners. */
	private final Set<ILabelProviderListener> labelProviderListeners = new IdentityHashSet<ILabelProviderListener>();

	/** Stores the UI update job. */
	private final UIJob updateUI = new UIJob("Update Model Decorators") {
		@Override
		public IStatus runInUIThread(IProgressMonitor monitor) {
			LabelProviderChangedEvent changedEvent = new LabelProviderChangedEvent(
					MarkerService.this);
			for (ILabelProviderListener listener : labelProviderListeners) {
				listener.labelProviderChanged(changedEvent);
			}
			if (markerView != null) {
				markerView.refresh();
			}
			return Status.OK_STATUS;
		}
	};

	/**
	 * List of invalid top-level elements to be refreshed by the constraint
	 * checker job.
	 */
	private final List<ITopLevelElement> invalidElements = new LinkedList<ITopLevelElement>();

	/** Stores the marker view. */
	private MarkerViewPart markerView;

	/** Stores the constraint checking job. */
	private final Job constraintCheckerJob = new Job(
			"Model Constraint Checker Job") {

		@Override
		protected IStatus run(IProgressMonitor monitor) {
			ITopLevelElement toBeRefreshed;
			// get next element to be refreshed
			synchronized (invalidElements) {
				if (invalidElements.isEmpty()) {
					return Status.OK_STATUS;
				}
				toBeRefreshed = invalidElements.remove(0);
			}
			while (toBeRefreshed != null) {
				refreshMarkers(toBeRefreshed);
				// get next element to be refreshed
				synchronized (invalidElements) {
					if (invalidElements.isEmpty()) {
						toBeRefreshed = null;
					} else {
						toBeRefreshed = invalidElements.remove(0);
					}
				}
			}
			return Status.OK_STATUS;
		}
	};

	/** Constructor. */
	public MarkerService() {
		for (ITopLevelElement element : IPersistencyService.INSTANCE
				.getTopLevelElements()) {
			invalidElements.add(element);
		}
		constraintCheckerJob.schedule();
		IPersistencyService.INSTANCE.addTopLevelElementListener(this);
	}

	/** Sets the current marker view. */
	public void setMarkerViewPart(MarkerViewPart view) {
		this.markerView = view;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IConstraintViolation<? extends EObject>> getViolations(
			EObject element) {
		ITopLevelElement top = IPersistencyService.INSTANCE
				.getTopLevelElementFor(element);
		if (top == null) {
			return Collections.emptyList();
		}
		return getCacheEntry(top).getCachedList(element);
	}

	/** {@inheritDoc} */
	@Override
	public ESeverity getHighestViolationSeverity(EObject element) {
		ITopLevelElement top = IPersistencyService.INSTANCE
				.getTopLevelElementFor(element);
		if (top == null) {
			return ESeverity.lowest();
		}
		return getCacheEntry(top).getHighestSeverity(element);
	}

	/** Accesses cache. */
	private CacheEntry getCacheEntry(ITopLevelElement element) {
		synchronized (violationCache) {
			CacheEntry result = violationCache.get(element);
			if (result == null) {
				result = new CacheEntry(element);
				violationCache.put(element, result);
			}
			return result;
		}
	}

	/** {@inheritDoc} */
	@Override
	public void refreshMarkers(ITopLevelElement element) {

		List<IConstraintViolation<? extends EObject>> checkResult = IConstraintCheckerService.INSTANCE
				.performAllConstraintChecksRecursively(element
						.getRootModelElement());
		synchronized (violationCache) {
			CacheEntry entry = getCacheEntry(element);
			entry.updateCacheEntries(checkResult);
		}

		updateUI.schedule();
	}

	/** {@inheritDoc} */
	@Override
	public void topLevelElementAdded(ITopLevelElement element) {
		doConstraintCheck(element);
	}

	/** {@inheritDoc} */
	@Override
	public void topLevelElementChanged(ITopLevelElement element) {
		doConstraintCheck(element);
	}

	/** {@inheritDoc} */
	@Override
	public void topLevelElementRemoved(ITopLevelElement element) {
		clearViolationCache(element);
	}

	/** {@inheritDoc} */
	@Override
	public void decorate(Object element, IDecoration decoration) {
		if (element instanceof EObject) {
			EObject modelElement = (EObject) element;
			ESeverity highest = getHighestViolationSeverity(modelElement);
			// TODO (FH): https://af3.fortiss.org/issues/388
			// use preferences settings here
			switch (highest) {
			case FATAL:
				// fatal uses error icon
			case ERROR:
				decoration.addOverlay(ESharedImages.ERROR_OVERLAY
						.getImageDescriptor());
				break;
			case WARNING:
				decoration.addOverlay(ESharedImages.WARNING_OVERLAY
						.getImageDescriptor());
				break;
			default:
				// ignore info and debug
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		IPersistencyService.INSTANCE.removeTopLevelElementListener(this);
	}

	/** {@inheritDoc} */
	@Override
	public boolean isLabelProperty(Object element, String property) {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public void addListener(ILabelProviderListener listener) {
		labelProviderListeners.add(listener);
	}

	/** {@inheritDoc} */
	@Override
	public void removeListener(ILabelProviderListener listener) {
		labelProviderListeners.remove(listener);
	}

	/** Schedules the given element for constraint checking. */
	private void doConstraintCheck(ITopLevelElement element) {
		synchronized (invalidElements) {
			invalidElements.add(element);
		}
		constraintCheckerJob.schedule();
	}

	/**
	 * Clears the cache from all elements associated with the given top-level
	 * element.
	 */
	private void clearViolationCache(ITopLevelElement element) {
		synchronized (violationCache) {
			violationCache.remove(element);
		}
	}

	/**
	 * Marker service cache entry.
	 */
	private static class CacheEntry {

		/** Stores the mapping from model elements to violation lists. */
		private final Map<EObject, List<IConstraintViolation<? extends EObject>>> violationsMap = new HashMap<EObject, List<IConstraintViolation<? extends EObject>>>();

		/** Stores the highest severity value for each cached element. */
		private final Map<EObject, ESeverity> highestSeverityMap = new HashMap<EObject, ESeverity>();

		/** Stores the top-level element. */
		private ITopLevelElement topElement;

		/** Constructor. */
		public CacheEntry(ITopLevelElement topElement) {
			this.topElement = topElement;
		}

		/** Returns the highest severity for the given element. */
		public ESeverity getHighestSeverity(EObject element) {
			ESeverity severity = highestSeverityMap.get(element);
			if (severity == null) {
				severity = ESeverity.lowest();
			}
			return severity;
		}

		/** Updates the cache entry. */
		public synchronized void updateCacheEntries(
				List<IConstraintViolation<? extends EObject>> newViolations) {
			clearCachedLists();
			// update cache entries
			for (IConstraintViolation<? extends EObject> violation : newViolations) {
				getCachedList(violation.getSource()).add(violation);
			}
			// fix cached list order
			for (EObject eo : violationsMap.keySet()) {
				List<IConstraintViolation<? extends EObject>> list = violationsMap
						.get(eo);
				if (list.isEmpty()) {
					highestSeverityMap.put(eo, ESeverity.lowest());
				} else {
					Collections.sort(list,
							IConstraintViolation.SEVERITY_COMPARATOR);
					highestSeverityMap.put(eo, list.get(0).getSeverity());
				}
			}
			// project severity to parent
			computeHighestSeverity(topElement.getRootModelElement());
		}

		/** Recursively projects highest severity from children to parents. */
		private ESeverity computeHighestSeverity(EObject element) {
			ESeverity severity = getHighestSeverity(element);
			for (EObject child : element.eContents()) {
				ESeverity childSeverity = computeHighestSeverity(child);
				if (ESeverity.getIntSeverity(severity) > ESeverity
						.getIntSeverity(childSeverity)) {
					severity = childSeverity;
				}
			}
			highestSeverityMap.put(element, severity);
			return severity;
		}

		/** Returns the cached list instance creating it if necessary. */
		private List<IConstraintViolation<? extends EObject>> getCachedList(
				EObject element) {
			List<IConstraintViolation<? extends EObject>> list = violationsMap
					.get(element);
			if (list == null) {
				list = new ArrayList<IConstraintViolation<? extends EObject>>();
				violationsMap.put(element, list);
			}
			return list;
		}

		/** Returns all violations of the given severity. */
		public List<IConstraintViolation<? extends EObject>> getAllViolationsWithSeverity(
				ESeverity severity) {
			List<IConstraintViolation<? extends EObject>> list = new ArrayList<IConstraintViolation<? extends EObject>>();
			for (EObject element : violationsMap.keySet()) {
				for (IConstraintViolation<? extends EObject> violation : violationsMap
						.get(element)) {
					if (violation.getSeverity() == severity) {
						list.add(violation);
					}
				}
			}
			return list;
		}

		/** Clears all cached lists. */
		private void clearCachedLists() {
			for (List<IConstraintViolation<? extends EObject>> list : violationsMap
					.values()) {
				list.clear();
			}
			for (EObject eo : highestSeverityMap.keySet()) {
				highestSeverityMap.put(eo, ESeverity.lowest());
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IConstraintViolation<? extends EObject>> getAllViolationsWithSeverity(
			ESeverity severity) {
		List<IConstraintViolation<? extends EObject>> list = new ArrayList<IConstraintViolation<? extends EObject>>();
		for (ITopLevelElement top : IPersistencyService.INSTANCE
				.getTopLevelElements()) {
			list.addAll(getCacheEntry(top).getAllViolationsWithSeverity(
					severity));
		}
		return list;
	}
}
